/*
 * S    (Sound)
 *
 * this class holds the sound code we use
 * in CS 5. It's one letter for ease of typing... .
 */
 
import java.applet.*;
import java.io.*;
import java.net.*;
import java.util.*;

import  javax.sound.sampled.AudioFileFormat;
import  javax.sound.sampled.AudioFileFormat.Type;
import  javax.sound.sampled.AudioFormat;
import  javax.sound.sampled.AudioInputStream;
import  javax.sound.sampled.AudioSystem;
import  javax.sound.sampled.Clip;
import  javax.sound.sampled.DataLine;
import  javax.sound.sampled.LineEvent;
import  javax.sound.sampled.LineListener;
import  javax.sound.sampled.LineUnavailableException;

class S
{
    
    public static void playSound(String filename)
    {
        try
        {
            URL testUrl = new URL("file:" + filename); // ..\\trippygaia1.mid");
            AudioClip testClip = Applet.newAudioClip(testUrl);
            testClip.play();
        } 
        catch (Exception e) 
        {
            H.pl("Could not play the requested sound..."); // e.printStackTrace();
        }
    }
    
    public static int DEBUG_LEVEL = 0;
    
    public static void play(SoundClip sc)
    {
        play(sc,1);
    }
    
    public static void play(SoundClip sc, int nLoops)
    {
        if (sc != null && sc.m_clip != null)
        {
            if ( !sc.m_clip.isOpen() )
            {
                try
                {
                    sc.m_clip.open(sc.af, sc.soundBuffer, 0, sc.soundBuffer.length);
                }
                catch (Exception e)
                {
                    e.printStackTrace();
                    return;
                }
            }
            
            // now it's open -- do we want to close it if it was open?
            // we won't for now...
            if (nLoops < 1) nLoops = 1;
            
            sc.m_clip.loop(nLoops-1);   // count from 1 instead of 0
            
            // wait until it's finished playing...
            while ( sc.m_clip.isRunning() )
            {
                H.pause(100);
            }
            
            // release the Line (whatever that means)
            sc.m_clip.close();
        }
        else
        {
            H.pl("No sound in the requested SoundClip");
        }
    }
    
    public static void writeToWav(SoundClip sc, String filename)
    {
        if (sc != null && sc.m_clip != null)
        {
            // first we create an input stream from the buffer...
            ByteArrayInputStream bais = new ByteArrayInputStream(sc.soundBuffer);
            AudioInputStream ais = new AudioInputStream(bais,sc.af,(long)(sc.soundBuffer.length/sc.af.getFrameSize()));
            File fileOut = new File(filename);
            try
            {
                AudioSystem.write(ais,AudioFileFormat.Type.WAVE,fileOut);
            }
            catch (Exception e)
            {
                e.printStackTrace();
            }
            
        }
        else
        {
            H.pl("No sound data to write to the file " + filename);
        }
        
        
        /* freeware to convert mp3 to wav and back:
         *   http://www.hitsquad.com/smm/programs/Mobsoft_Mp3toWav/
         */
        
        /* lots of code !! at http://www.musicdsp.org/files/(pink.txt)
         *
         */

        
    }
    
    public static SoundClip loadClip(String filename)
    {
        
        int         m_nLoopCount = 1;
        Clip           m_clip = null;
        AudioFormat        af = null;
        AudioInputStream  ais = null;
        AudioFileFormat   aff = null;
        byte[]    soundBuffer = null;
        
        try
        {
            File f = new File(filename);
            
            if (S.DEBUG_LEVEL == 1)
            { 
                aff = AudioSystem.getAudioFileFormat(f);
                H.pl("The audio file format is " + aff);
            }
            
            ais = AudioSystem.getAudioInputStream(f);
            
            if (S.DEBUG_LEVEL == 1) 
            {
                // let's find out what's up here...
                H.pl("ais is " + ais);
                H.pl("the # of available bytes w/o blocking is " + ais.available());
                H.pl("the frame length of the whole stream is " + ais.getFrameLength());
                H.pl("the markSupported status of this stream is " + ais.markSupported());
                H.pl();
            }
        }
        catch (Exception e)
        {
            e.printStackTrace();
            return new SoundClip(null,null);
        }
        
        if (ais != null)
        {
            af = ais.getFormat();

            if (S.DEBUG_LEVEL == 1)
            {
                // for fun. let's print what's going on in this format...
                H.pl("The format is  " + af);
                H.pl("the number of bytes per frame is " + af.getFrameSize());
                H.pl();
            }
            
            // let's copy all of the audio input stream's data to a buffer!
            int frameLength = (int)(ais.getFrameLength());
            int frameSize = af.getFrameSize();
            int bufferSize = frameLength * frameSize;
            
            soundBuffer = new byte[bufferSize];
            int nReadBytes = 0;
            
            try
            {
                // I think this should be a different read command or a 
                // loop or something like that!
                nReadBytes = ais.read(soundBuffer,0,soundBuffer.length);
            }
            catch (Exception e)
            {
                e.printStackTrace();
                return new SoundClip(null,null);
            }
            
            if (S.DEBUG_LEVEL == 1)
            {
                H.pl(" the buffer size is " + bufferSize);
                H.pl("number of read bytes was " + nReadBytes);
            }
                  
            DataLine.Info info = new DataLine.Info(Clip.class, af);
            
            if (S.DEBUG_LEVEL == 1)
            {
                H.pl("info is " + info);
            }
            
            try
            {
                m_clip = (Clip)AudioSystem.getLine(info);
                
                // this next line of code opens the clip for playing the data from the
                // audio input stream BUT we want to use the alternative open
                // provided for clips that allows us to write all the data at once!
                // m_clip.open(ais);
                
                // this will be done in the play method
                // m_clip.open(af, soundBuffer, 0, soundBuffer.length);
                
                // this is for more advanced stuff...
                // m_clip.addLineListener(new SListener()); 
            }
            catch (Exception e)
            {
                e.printStackTrace();
                return new SoundClip(null,null);
            }
        }
        else
        {
            System.out.println("ClipPlayer.<init>(): can't get data from file " + filename);
        }
        
        return new SoundClip(m_clip,soundBuffer,af);
    }

}


class SoundClip 
{
    
    Clip        m_clip;
    byte[]      soundBuffer;
    AudioFormat af;
    int         soundBufferLength;

    public SoundClip(Clip c, byte[] ba, AudioFormat af_in)
    {
        this.m_clip = c;
        this.soundBuffer = ba;
        this.af = af_in;
    }
    
    public byte[] getBytes()
    {
        return soundBuffer;
    }
    
    // get rid of this dummyconstructor when there's time
    public SoundClip(Clip c, byte[] ba)
    {
        this.m_clip = null;
        this.soundBuffer = null;
        this.af = null;
    }
    
}


//class SListenerOld implements LineListener
//{
//    public void update(LineEvent event)
//    {
//        if (event.getType().equals(LineEvent.Type.STOP))
//        {
//            H.pl("STOP Line event");
////            if (m_clip != null)
////                m_clip.close();
//        }
//        else if (event.getType().equals(LineEvent.Type.CLOSE))
//        {
//            H.pl("CLOSE Line event");
//            /*
//             *  There is a bug in the jdk1.3.0.
//             *  It prevents correct termination of the VM.
//             *  So we have to exit ourselves.
//             */
//            //System.exit(0);
////            if (m_clip != null)
////                m_clip.close();
//        }
//        else if (event.getType().equals(LineEvent.Type.OPEN))
//        {
//            H.pl("OPEN Line Event");
//        }
//        else
//        {
//            H.pl("Line event type: " + event.getType());
//        }
//
//    }
//}


